import tele.meter
import torchnet.meter
import tele.console.views


class Sink(tele.Sink):
    def write(self, step_num, rendered):
        print('[{:4d}] '.format(step_num) + ', '.join(rendered))


class Conf(tele.Conf):
    def make_auto_view(self, meter_name, meter):
        if isinstance(meter, torchnet.meter.AverageValueMeter) \
                or isinstance(meter, torchnet.meter.TimeMeter) \
                or isinstance(meter, tele.meter.ValueMeter):
            return tele.console.views.KeyValue([meter_name])
        return None

    def build(self, view_list):
        cell_list = [view.build() for view in view_list]
        return Sink(cell_list)



import unittest

from tele import Telemetry
from tele.meter import ValueMeter
import tele.showoff
import tele.showoff.views
import pyshowoff


class DummyClient(pyshowoff.Client):
    def __init__(self, callback=None):
        super().__init__('dummy:1337', True)
        self.callback = callback

    def request(self, method, path, data=None):
        if self.callback:
            return self.callback(method, path, data)
        return None


class TestShowoffOutput(unittest.TestCase):
    def test_normal(self):
        patch_data = {}

        def callback(method, path, data):
            if method.lower() == 'post' and path == '/api/v2/frames':
                return {'data': {'id': '100'}}
            if method.lower() == 'patch' and path == '/api/v2/frames/100':
                patch_data.clear()
                patch_data.update(data)
                return None
            self.fail('unexpected Showoff request')

        client = DummyClient(callback)
        notebook = pyshowoff.Notebook(client, 1)
        t = Telemetry({'val': ValueMeter()})
        t['val'].set_value(42)
        t.sink(tele.showoff.Conf(notebook), [
            tele.showoff.views.Text(['val'], 'Value'),
        ])
        t.step()
        self.assertEqual(patch_data, {'data': {
            'id': '100',
            'type': 'frames',
            'attributes': {'type': 'text', 'content': {'body': '42'}}
        }})

    def test_histogram(self):
        patch_data = {}

        def callback(method, path, data):
            if method.lower() == 'post' and path == '/api/v2/frames':
                return {'data': {'id': '100'}}
            if method.lower() == 'patch' and path == '/api/v2/frames/100':
                patch_data.clear()
                patch_data.update(data)
                return None
            self.fail('unexpected Showoff request')

        client = DummyClient(callback)
        notebook = pyshowoff.Notebook(client, 1)
        t = Telemetry({'val': ValueMeter()})
        t['val'].set_value([0.9, 1, 1.1, 2, 2, 3, 4, 3])
        t.sink(tele.showoff.Conf(notebook), [
            tele.showoff.views.Histogram(['val'], 'Value', bins=10, extent=[0, 10], x_title='val'),
        ])
        t.step()
        self.assertEqual(patch_data, {'data': {
            'id': '100',
            'type': 'frames',
            'attributes': {
                'type': 'vegalite',
                'content': {'body': {
                    'width': 420,
                    'height': 250,
                    'data': {'values': [
                        {'x': '0.50', 'y': 1}, {'x': '1.50', 'y': 2}, {'x': '2.50', 'y': 2},
                        {'x': '3.50', 'y': 2}, {'x': '4.50', 'y': 1}, {'x': '5.50', 'y': 0},
                        {'x': '6.50', 'y': 0}, {'x': '7.50', 'y': 0}, {'x': '8.50', 'y': 0},
                        {'x': '9.50', 'y': 0}
                    ]},
                    'mark': 'bar',
                    'encoding': {
                        'x': {'field': 'x', 'type': 'ordinal', 'sort': False,
                              'axis': {'title': 'val', 'labelAngle': 0}},
                        'y': {'field': 'y', 'type': 'quantitative', 'axis': {'title': 'count'}}
                    }
                }}
            }
        }})



class Telemetry():
    def __init__(self, meters):
        self.meters = meters
        self.sinks = []
        self.step_num = 0

    def sink(self, conf, view_list=None, auto_view=False):
        if view_list is None:
            view_list = []
        if auto_view:
            meters_with_views = []
            for view in view_list:
                meters_with_views.extend(view.meter_names)
            meters_without_views = set(self.meters.keys()) - set(meters_with_views)
            for meter_name in sorted(meters_without_views):
                view = conf.make_auto_view(meter_name, self.meters[meter_name])
                if view is not None:
                    view_list.append(view)
        sink = conf.build(view_list)
        self.sinks.append(sink)
        return sink

    def step(self):
        for sink in self.sinks:
            rendered = sink.render_all(self.step_num, self.meters)
            sink.write(self.step_num, rendered)
        for meter_name, meter in self.meters.items():
            if not hasattr(meter, 'skip_reset') or not meter.skip_reset:
                meter.reset()
        self.step_num += 1

    def __getitem__(self, meter_name):
        return self.meters[meter_name]


# Descendants of this class are for internal use. They are built by their
# corresponding Conf.
class Sink:
    def __init__(self, cell_list):
        self.cell_list = cell_list

    def render_all(self, step_num, meters):
        rendered = []
        for cell in self.cell_list:
            cell_meters = [meters[mn] for mn in cell.meter_names]
            rendered.append(cell.render(step_num, cell_meters))
        return rendered

    def write(self, step_num, rendered):
        pass


# Descendants of this class are part of the public API, and are used to
# configure an output.
class Conf:
    def make_auto_view(self, meter_name, meter):
        return None

    def build(self, view_list):
        raise NotImplementedError()


# Descendants of this class are for internal use. They are built by
# their corresponding View.
class Cell:
    def __init__(self, meter_names):
        self.meter_names = meter_names

    def render(self, step_num, meters):
        raise NotImplementedError()


# Descendants of this class are part of the public API, and are used to
# specify how groups of one or more meters are displayed.
class View:
    def __init__(self, meter_names):
        self.meter_names = meter_names

    def build(self, *args):
        raise NotImplementedError()



import tele
import torchnet.meter


class _KeyValueCell(tele.Cell):
    def render(self, step_num, meters):
        meter_name = self.meter_names[0]
        meter = meters[0]
        value = meter.value()
        if isinstance(meter, torchnet.meter.AverageValueMeter):
            (mean, std) = value
            value_str = u'{:0.4f}\u00b1{:0.4f}'.format(mean, std)
        elif isinstance(value, float):
            value_str = '{:0.4f}'.format(value)
        else:
            value_str = str(value)
        return '='.join([meter_name, value_str])


class KeyValue(tele.View):
    def build(self):
        return _KeyValueCell(self.meter_names)



import tele
from io import StringIO, BytesIO
import base64
from PIL import Image
import torchvision.transforms as transforms
import torchnet.meter
import numpy as np


class Cell(tele.Cell):
    def __init__(self, meter_names, frame):
        super().__init__(meter_names)
        self.frame = frame

    def render(self, step_num, meters):
        raise NotImplementedError()


class View(tele.View):
    def __init__(self, meter_names, frame_title='untitled'):
        super().__init__(meter_names)
        self.frame_title = frame_title

    def build(self, frame):
        raise NotImplementedError()


class _TextCell(Cell):
    def render(self, step_num, meters):
        stream = StringIO()
        for meter in meters:
            stream.write(str(meter.value()))
        self.frame.text(stream.getvalue())


class Text(View):
    def build(self, frame):
        return _TextCell(self.meter_names, frame)


class _InspectCell(Cell):
    def __init__(self, meter_names, frame, flatten):
        super().__init__(meter_names, frame)
        self.flatten = flatten

    def render(self, step_num, meters):
        items = []
        for meter_name, meter in zip(self.meter_names, meters):
            value = meter.value()
            if self.flatten and hasattr(value, 'items'):
                for k, v in value.items():
                    key = meter_name + '.' + k
                    str_value = str(v)
                    items.append((key, str_value))
            else:
                key = meter_name
                str_value = str(value)
                items.append((key, str_value))

        stream = StringIO()
        stream.write('<table style="width: 100%; table-layout: fixed;">')
        for key, str_value in items:
            stream.write('<tr><td>')
            stream.write(key)
            stream.write('</td></tr><tr><td><pre>')
            stream.write(str_value)
            stream.write('</pre></td></tr>')
        stream.write('</table>')
        self.frame.html(stream.getvalue())


class Inspect(View):
    def __init__(self, meter_names, frame_title, flatten=False):
        super().__init__(meter_names, frame_title)
        self.flatten = flatten

    def build(self, frame):
        return _InspectCell(self.meter_names, frame, self.flatten)


class _GraphvizCell(Cell):
    def render(self, step_num, meters):
        meter = meters[0]
        value = meter.value()
        svg_bytes = value.pipe(format='svg')
        b64_str = base64.b64encode(svg_bytes).decode('utf-8')
        img_tag_template = '<img style="width: 100%;" src=data:image/svg+xml;base64,{}>'
        self.frame.html(img_tag_template.format(b64_str))


class Graphviz(View):
    def build(self, frame):
        return _GraphvizCell(self.meter_names, frame)


class _LineGraphCell(Cell):
    def __init__(self, meter_names, frame):
        super().__init__(meter_names, frame)
        self.xs = []
        self.yss = None

    def render(self, step_num, meters):
        if self.yss is None:
            self.yss = [[] for _ in meters]
        self.xs.append(step_num)
        series_names = []
        for i, meter in enumerate(meters):
            value = meter.value()
            if isinstance(meter, torchnet.meter.AverageValueMeter):
                value = value[0]
            self.yss[i].append(value)
            series_names.append(self.meter_names[i])
        self.frame.line_graph(self.xs, self.yss, series_names=series_names)


class LineGraph(View):
    def build(self, frame):
        return _LineGraphCell(self.meter_names, frame)


class _HistogramCell(Cell):
    def __init__(self, meter_names, frame, bins, extent, x_title):
        super().__init__(meter_names, frame)
        self.bins = bins
        self.extent = extent
        self.x_title = x_title

    def render(self, step_num, meters):
        vals = []
        for meter in meters:
            vals.extend(meter.value())
        ys, xs = np.histogram(vals, bins=self.bins, range=self.extent)

        xs = [float(x1 + x2) / 2 for x1, x2 in zip(xs, xs[1:])]

        values = [{'x': '{:0.2f}'.format(x), 'y': int(y)} for x, y in zip(xs, ys)]

        spec = {
            'width': 420,
            'height': 250,
            'data': {'values': values},
            'mark': 'bar',
            'encoding': {
                'x': {
                    'field': 'x',
                    'type': 'ordinal',
                    'sort': False,
                    'axis': {'title': self.x_title, 'labelAngle': 0},
                },
                'y': {
                    'field': 'y',
                    'type': 'quantitative',
                    'axis': {'title': 'count'},
                }
            }
        }
        self.frame.vegalite(spec)


class Histogram(View):
    def __init__(self, meter_names, frame_title, bins=10, extent=None, x_title='x'):
        super().__init__(meter_names, frame_title)
        self.bins = bins
        self.extent = extent
        self.x_title = x_title

    def build(self, frame):
        return _HistogramCell(self.meter_names, frame, self.bins, self.extent, self.x_title)


class _ImageCell(Cell):
    def __init__(self, meter_names, frame, images_per_row):
        super().__init__(meter_names, frame)
        self.images_per_row = images_per_row

    def render(self, step_num, meters):
        images = []
        value = None
        for meter in meters:
            value = meter.value()
            if isinstance(value, list):
                images.extend(value)
            else:
                images.append(value)

        if self.images_per_row is None:
            img_tag_template = '<img src=data:image/png;base64,{}>'
        else:
            width = '{:0.2f}%'.format(100 / self.images_per_row)
            img_tag_template = '<img style="width: ' + width + ';" src=data:image/png;base64,{}>'

        stream = StringIO()
        stream.write(
            '<div style="background-image: linear-gradient(45deg, #808080 25%, transparent 25%), '
            + 'linear-gradient(-45deg, #808080 25%, transparent 25%), '
            + 'linear-gradient(45deg, transparent 75%, #808080 75%), '
            + 'linear-gradient(-45deg, transparent 75%, #808080 75%); '
            + 'background-size: 20px 20px; '
            + 'background-position: 0 0, 0 10px, 10px -10px, -10px 0px;">')
        for img in images:
            if not isinstance(img, Image.Image):
                img = transforms.ToPILImage()(value.cpu())
            buf = BytesIO()
            img.save(buf, format='PNG')
            b64_str = base64.b64encode(buf.getvalue()).decode('utf-8')
            stream.write(img_tag_template.format(b64_str))
        stream.write('</div>')
        self.frame.html(stream.getvalue())


class Images(View):
    def __init__(self, meter_names, frame_title, images_per_row=None):
        super().__init__(meter_names, frame_title)
        self.images_per_row = images_per_row

    def build(self, frame):
        return _ImageCell(self.meter_names, frame, self.images_per_row)



import unittest

from tele.meter import ValueMeter


class TestValueMeter(unittest.TestCase):
    def test_set_value(self):
        meter = ValueMeter()
        self.assertIsNone(meter.value())
        meter.set_value(42)
        self.assertEqual(meter.value(), 42)



import tele.folder.views
import os


class Conf(tele.Conf):
    def __init__(self, dir_path):
        super().__init__()
        self.dir_path = dir_path

    def make_auto_view(self, meter_name, meter):
        return None

    def build(self, view_list):
        if not os.path.exists(self.dir_path):
            os.makedirs(self.dir_path)
        cell_list = [view.build(self.dir_path) for view in view_list]
        return tele.Sink(cell_list)



import unittest

from tele.meter import ListMeter


class TestValueMeter(unittest.TestCase):
    def test_add(self):
        meter = ListMeter()
        self.assertListEqual(meter.value(), [])
        meter.add(3)
        meter.add(7)
        self.assertListEqual(meter.value(), [3, 7])

    def test_add_all(self):
        meter = ListMeter()
        self.assertListEqual(meter.value(), [])
        meter.add_all([1, 2])
        meter.add_all([3, 4])
        self.assertListEqual(meter.value(), [1, 2, 3, 4])



import requests
import json
from threading import Thread


class Client:
    def __init__(self, netloc, always_sync=False):
        self.base_url = 'http://' + netloc
        self.session = requests.Session()
        self.session.headers.update({
            'content-type': 'application/json'
        })
        self.always_sync = always_sync

    def request(self, method, path, data=None):
        if data is not None:
            data = json.dumps(data)
        req = requests.Request(method, self.base_url + path, data=data)
        prepped = self.session.prepare_request(req)
        r = self.session.send(prepped)
        return r.json()

    def async_request(self, method, path, data=None):
        if self.always_sync:
            self.request(method, path, data)
        else:
            def do_request():
                self.request(method, path, data)
            Thread(target=do_request, daemon=False).start()

    def new_notebook(self, title):
        data = {
            'data': {
                'type': 'notebooks',
                'attributes': {'title': title},
            },
        }
        res = self.request('post', '/api/v2/notebooks', data)
        return Notebook(self, res['data']['id'])


class Notebook:
    def __init__(self, client, notebook_id):
        self.client = client
        self.id = notebook_id

    def new_tag(self, name):
        data = {
            'data': {
                'type': 'tags',
                'attributes': {'name': name},
                'relationships': {
                    'notebook': {
                        'data': {'type': 'notebooks', 'id': self.id},
                    },
                },
            },
        }
        res = self.client.request('post', '/api/v2/tags', data)
        return Tag(self.client, res['data']['id'])

    def new_frame(self, title, bounds=None):
        data = {
            'data': {
                'type': 'frames',
                'attributes': {'title': title},
                'relationships': {
                    'notebook': {
                        'data': {'type': 'notebooks', 'id': self.id},
                    },
                },
            },
        }
        if bounds is not None:
            data['data']['attributes'].update(bounds)
        res = self.client.request('post', '/api/v2/frames', data)
        return Frame(self.client, res['data']['id'])


class Tag:
    def __init__(self, client, tag_id):
        self.client = client
        self.id = tag_id


class Frame:
    def __init__(self, client, frame_id):
        self.client = client
        self.id = frame_id

    def set_title(self, title):
        data = {
            'data': {
                'id': self.id,
                'type': 'frames',
                'attributes': {'title': title},
            },
        }
        self.client.request('patch', '/api/v2/frames/' + self.id, data)

    def set_content(self, content_type, content_body):
        data = {
            'data': {
                'id': self.id,
                'type': 'frames',
                'attributes': {
                    'type': content_type,
                    'content': {'body': content_body},
                },
            },
        }
        self.client.async_request('patch', '/api/v2/frames/' + self.id, data)

    def vega(self, spec):
        self.set_content('vega', spec)

    def vegalite(self, spec):
        self.set_content('vegalite', spec)

    def text(self, message):
        self.set_content('text', message)

    def html(self, html):
        self.set_content('html', html)

    def progress(self, current_value, max_value):
        percentage = min(100 * current_value / max_value, 100)
        html = """<div class="progress">
            <div class="progress-bar" role="progressbar"
             aria-valuenow="{percentage:0.2f}" aria-valuemin="0" aria-valuemax="100"
             style="width: {percentage:0.2f}%; min-width: 40px;"
            >
                {percentage:0.2f}%
            </div>
        </div>""".format(percentage=percentage)
        self.html(html)

    def line_graph(self, xss, yss, series_names=None, x_title=None, y_title=None,
                   y_axis_min=None, y_axis_max=None):
        if not isinstance(xss[0], list):
            xss = [xss] * len(yss)

        show_legend = True
        if series_names is None:
            show_legend = False
            series_names = ['series_{:03d}'.format(i) for i in range(len(xss))]

        min_x = float('inf')
        max_x = -float('inf')
        min_y = float('inf')
        max_y = -float('inf')
        tables = []
        marks = []
        for i, xs in enumerate(xss):
            marks.append({
                'type': 'line',
                'from': {'data': 'table_{:03d}'.format(i)},
                'properties': {
                    'enter': {
                        'x': {'scale': 'x', 'field': 'x'},
                        'y': {'scale': 'y', 'field': 'y'},
                        'stroke': {'scale': 'c', 'value': series_names[i]},
                    }
                },
            })

            points = []
            for j, x in enumerate(xs):
                y = yss[i][j]
                min_x = min(x, min_x)
                max_x = max(x, max_x)
                min_y = min(y, min_y)
                max_y = max(y, max_y)
                points.append({'x': x, 'y': y})
            tables.append(points)

        data = []
        for i, table in enumerate(tables):
            data.append({
                'name': 'table_{:03d}'.format(i),
                'values': table
            })

        spec = {
            'width': 370,
            'height': 250,
            'data': data,
            'scales': [
                {
                    'name': 'x',
                    'type': 'linear',
                    'range': 'width',
                    'domainMin': min_x,
                    'domainMax': max_x,
                    'nice': True,
                    'zero': False,
                }, {
                    'name': 'y',
                    'type': 'linear',
                    'range': 'height',
                    'domainMin': y_axis_min or min_y,
                    'domainMax': y_axis_max or max_y,
                    'nice': True,
                    'zero': False,
                }, {
                    'name': 'c',
                    'type': 'ordinal',
                    'range': 'category10',
                    'domain': series_names,
                }
            ],
            'axes': [
                {'type': 'x', 'scale': 'x', 'title': x_title},
                {'type': 'y', 'scale': 'y', 'title': y_title, 'grid': True},
            ],
            'marks': marks,
        }

        if show_legend:
            spec['legends'] = [{'fill': 'c'}]

        self.vega(spec)



from io import StringIO
import numpy as np
from scipy.stats import norm


class Meter:
    def __init__(self, skip_reset):
        self.reset()
        self.skip_reset = skip_reset

    def reset(self):
        pass

    def value(self):
        return None


class ValueMeter(Meter):
    def __init__(self, skip_reset=False):
        self._value = None
        super().__init__(skip_reset)

    def set_value(self, value):
        self._value = value

    def reset(self):
        self._value = None

    def value(self):
        return self._value


class ListMeter(Meter):
    def __init__(self, skip_reset=False):
        self._list = None
        super().__init__(skip_reset)

    def add(self, value):
        self._list.append(value)

    def add_all(self, values):
        self._list.extend(values)

    def reset(self):
        self._list = []

    def value(self):
        return self._list


class StringBuilderMeter(Meter):
    def __init__(self, skip_reset=False):
        self._value = None
        super().__init__(skip_reset)

    def add(self, part):
        self._value.write(part)

    def add_line(self, line):
        self.add(line)
        self.add('\n')

    def reset(self):
        self._value = StringIO()

    def value(self):
        return self._value.getvalue()


class MaxValueMeter(Meter):
    def __init__(self, skip_reset=False):
        self._value = None
        super().__init__(skip_reset)

    def add(self, new_value):
        if self._value is not None and new_value <= self._value:
            return False

        self._value = new_value
        return True

    def reset(self):
        self._value = None

    def value(self):
        return self._value


class SumMeter(Meter):
    def __init__(self, initial_value=0, skip_reset=False):
        self.initial_value = initial_value
        self._value = None
        super().__init__(skip_reset)

    def add(self, value):
        self._value += value

    def value(self):
        return self._value

    def reset(self):
        self._value = self.initial_value


class MedianValueMeter(Meter):
    def __init__(self, skip_reset=False):
        self.values = []
        super().__init__(skip_reset)

    def add(self, value):
        self.values.append(value)

    def value(self):
        data = np.asarray(self.values)

        median = np.median(data)

        # Calculate median absolute deviation (assumes data is normally distributed)
        k = norm.ppf(0.75)
        mad = np.median(np.fabs(data - median) / k)

        return median, mad

    def reset(self):
        self.values.clear()



import unittest

from tele import Telemetry
from tele.meter import ValueMeter
import tele.folder
import tele.folder.views
from tempfile import TemporaryDirectory
import os
import json


class TestFolderOutput(unittest.TestCase):
    def test_normal(self):
        t = Telemetry({'val': ValueMeter()})
        t['val'].set_value(42)
        with TemporaryDirectory() as tmpdir:
            t.sink(tele.folder.Conf(tmpdir), [
                tele.folder.views.JSON(['val'], 'test.json')
            ])
            t.step()

            file_path = os.path.join(tmpdir, 'test.json')
            self.assertTrue(os.path.isfile(file_path), '{} is not a file'.format(file_path))
            with open(file_path, 'r') as f:
                data = json.load(f)
                self.assertEqual(data, {'val': 42})



import unittest

from tele import Telemetry
from tele.meter import ValueMeter
import tele.console
import tele.console.views

from tests import captured_output


class TestConsoleOutput(unittest.TestCase):
    def test_normal(self):
        t = Telemetry({'val': ValueMeter()})
        t['val'].set_value(42)
        t.sink(tele.console.Conf(), [
            tele.console.views.KeyValue(['val'])
        ])
        with captured_output() as (out, err):
            t.step()
            self.assertEqual(out.getvalue().strip(), '[   0] val=42')

    def test_auto_view(self):
        t = Telemetry({'val': ValueMeter()})
        t['val'].set_value(42)
        t.sink(tele.console.Conf(), auto_view=True)
        with captured_output() as (out, err):
            t.step()
            self.assertEqual(out.getvalue().strip(), '[   0] val=42')



from setuptools import setup, find_packages

setup(
    name='tele',
    version='0.1.0a5',
    packages=find_packages(),
    author='Aiden Nibali',
    description='Telemetry for PyTorch',
    test_suite='tests',
    install_requires=[
        'graphviz',
        'h5py',
        'numpy',
        'Pillow',
        'requests',
        'torch',
        'torchnet',
        'torchvision',
    ],
)



import tele
import tele.meter
import torchnet.meter
import tele.showoff.views

Sink = tele.Sink


class Conf(tele.Conf):
    def __init__(self, notebook):
        super().__init__()
        self.notebook = notebook

    def make_auto_view(self, meter_name, meter):
        if isinstance(meter, tele.meter.StringBuilderMeter):
            return tele.showoff.views.Text(meter_name)
        if isinstance(meter, torchnet.meter.AverageValueMeter) \
                or isinstance(meter, torchnet.meter.TimeMeter):
            return tele.showoff.views.LineGraph(meter_name)
        return None

    @staticmethod
    def _calc_frame_bounds(index):
        cell_width = 480
        cell_height = 308
        cell_cols = 4

        r, c = divmod(index, cell_cols)

        return {
            'x': c * cell_width,
            'y': r * cell_height,
            'width': cell_width,
            'height': cell_height
        }

    def build(self, view_list):
        cell_list = []
        for i, view in enumerate(view_list):
            bounds = self._calc_frame_bounds(i)
            cell = view.build(self.notebook.new_frame(view.frame_title, bounds))
            cell_list.append(cell)
        return Sink(cell_list)



import tele
import os
import json
import h5py


class Cell(tele.Cell):
    def __init__(self, meter_names, dir_path, filename_template):
        super().__init__(meter_names)
        self.dir_path = dir_path
        self.filename_template = filename_template

    def build_path(self, step_num):
        return os.path.join(self.dir_path, self.filename_template.format(step_num))

    def render(self, step_num, meters):
        raise NotImplementedError()


class View(tele.View):
    def __init__(self, meter_names, file_name):
        super().__init__(meter_names)
        self.file_name = file_name

    def build(self, dir_path):
        raise NotImplementedError()


class _JSONCell(Cell):
    def render(self, step_num, meters):
        path = self.build_path(step_num)
        values = {}
        for meter_name, meter in zip(self.meter_names, meters):
            values[meter_name] = meter.value()
        with open(path, 'w') as f:
            json.dump(values, f)


class JSON(View):
    def __init__(self, meter_names, file_name='metrics_{:04d}.json'):
        super().__init__(meter_names, file_name)

    def build(self, dir_path):
        return _JSONCell(self.meter_names, dir_path, self.file_name)


class _GrowingJSONCell(Cell):
    def render(self, step_num, meters):
        file_path = self.build_path(step_num)
        if os.path.isfile(file_path):
            with open(file_path, 'r') as f:
                values = json.load(f)
        else:
            values = {}
        for meter_name, meter in zip(self.meter_names, meters):
            if meter_name not in values:
                values[meter_name] = []
            values[meter_name].append(meter.value())
        with open(file_path, 'w') as f:
            json.dump(values, f)


class GrowingJSON(View):
    def __init__(self, meter_names, file_name='metrics.json'):
        super().__init__(meter_names, file_name)

    def build(self, dir_path):
        return _GrowingJSONCell(self.meter_names, dir_path, self.file_name)


class _HDF5Cell(Cell):
    def __init__(self, meter_names, dir_path, filename_template, path_map):
        super().__init__(meter_names, dir_path, filename_template)
        self.path_map = path_map

    def render(self, step_num, meters):
        file_path = self.build_path(step_num)
        with h5py.File(file_path, 'w') as f:
            for meter_name, meter in zip(self.meter_names, meters):
                if meter_name in self.path_map:
                    h5_path = self.path_map[meter_name]
                else:
                    h5_path = meter_name
                f.create_dataset(h5_path, data=meter.value())


class HDF5(View):
    def __init__(self, meter_names, file_name='data_{:04d}.h5', path_map=None):
        super().__init__(meter_names, file_name)
        if path_map is None:
            path_map = {}
        self.path_map = path_map

    def build(self, dir_path):
        return _HDF5Cell(self.meter_names, dir_path, self.file_name, self.path_map)
